"""Utilities for working with Python packages"""
import sys
import os
import inspect

def add_package_to_path():
    """Programmatically determine the most likely root of the current running program, add its parent to the path, and return the root folder name"""
    # Define the list of common package root files and folders with lowercase
    package_root_items = ["src", "tests", "templates", "docs", "dist", "build", "readme.md", "license.txt", ".gitignore", "pyproject.toml", "requirements.txt", "poetry.lock", "setup.py", "manifest.in", ".editorconfig"]

    # Initialize a dictionary to store the count of package root items found in each directory
    item_counts = {}

    # Traverse upward from the current directory, counting the instances of package root items found in each directory
    current_dir = os.getcwd()
    while current_dir != os.path.dirname(current_dir):
        # Count the instances of package root items found in the current directory
        current_item_counts = {item: 0 for item in package_root_items}
        for item in os.listdir(current_dir):
            item_lower = item.lower()
            if item_lower in current_item_counts:
                current_item_counts[item_lower] += 1

        # Add the counts for the current directory to the overall counts dictionary
        for item, count in current_item_counts.items():
            if count > 0:
                item_counts.setdefault(item, {})[current_dir] = count

        # Move up one directory and continue counting
        current_dir = os.path.dirname(current_dir)

    # Find the directory with the most package root items found, and add its parent to the path
    max_item_count = 0
    max_item_count_dir = ""
    for item, counts in item_counts.items():
        for directory, count in counts.items():
            if count > max_item_count or (count == max_item_count and len(directory) > len(max_item_count_dir)):
                max_item_count = count
                max_item_count_dir = directory

    if max_item_count > 0:
        package_root_dir = os.path.dirname(max_item_count_dir)
        os.environ["PATH"] += os.pathsep + package_root_dir
        package_root_name = os.path.basename(max_item_count_dir)
        os.environ["PATH"] += os.pathsep + package_root_name
        print(f'PATH: {os.environ["PATH"]}')
        return package_root_dir, package_root_name
    else:
        print("Could not find package root directory")
        return None

def import_relative(package_root_name, module_path, import_name, alias=None):
    """Import a relative library dynamically.  Accepts 'package_root_name' from 'add_package_to_path()' function"""
    module_name = f"{package_root_name}.{module_path}"
    print(f'module_name: {module_name}')
    module = __import__(module_name, fromlist=[import_name])

    # Get the imported object
    obj = getattr(module, import_name)

    # Determine the globals of the calling program and add any alias provided to it
    caller_globals = inspect.stack()[1][0].f_globals
    if alias:
        caller_globals[alias] = obj
    else:
        caller_globals[import_name] = obj

    # Return the imported object
    # return obj



"""Password-related utilities"""
import boto3
import os
from botocore.exceptions import ClientError
import keyring

def get_password(password_method, password_key, account_name=None, access_key=None, secret_key=None, endpoint_url=None, password_path=None, encoding='utf-8'):
    """Retrieve password based on method provided"""
    if password_method == 'keyring':
        try:
            response = keyring.get_password(account_name, password_key)
        except Exception as e:
            print(f"Error retrieving secret: {e}")
            return None
        else:
            secret_value = response
            return secret_value
        
    elif password_method == 'ssm':
        access_key = os.environ.get('AWS_SSM_ACCESS_KEY_ID') or access_key
        secret_key = os.environ.get('AWS_SSM_SECRETACCESS_KEY_ID') or secret_key
        ssm_endpoint_url = os.environ.get('AWS_SSM_ENDPOINT_URL') or endpoint_url
        ssm_password_path = os.environ.get('AWS_SSM_PASSWORD_PATH') or password_path

        if not all([access_key, secret_key, ssm_endpoint_url, ssm_password_path]):
            raise ValueError('One or more required environment variables is not set')

        sts_client = boto3.client('sts', aws_access_key_id=access_key, aws_secret_access_key=secret_key)

        response = sts_client.get_session_token()
        access_key_id = response['Credentials']['AccessKeyId']
        secret_access_key = response['Credentials']['SecretAccessKey']
        session_token = response['Credentials']['SessionToken']

        client = boto3.client(
            'ssm',
            # endpoint_url='https://ssm.us-west-1.amazonaws.com',
            # region_name='us-west-1',
            endpoint_url=ssm_endpoint_url,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            aws_session_token=session_token
        )

        # password_path = '/flat_file_loader/passwords/dev'
        password_path = ssm_password_path
        # parameter_name = f"{account_name}_{password_key}"
        parameter_name = password_key
        if password_path is not None and password_path != '':
            # parameter_name = f"{password_path}/{account_name}_{password_key}"
            parameter_name = f"{password_path}/{password_key}"
        try:
            response = client.get_parameter(Name=parameter_name, WithDecryption=True)
        except ClientError as e:
            print(f"Error retrieving secret: {e}")
            return None
        else:
            secret_value = response['Parameter']['Value']
            supported_encodings = ['utf-8', 'ascii', 'latin-1', 'utf-16']
            if encoding not in supported_encodings:
                print(f"Unsupported encoding: {encoding}")
                return None
            if isinstance(secret_value, bytes):
                secret_value = secret_value.decode(encoding)
        
        return secret_value
    
    elif password_method == 'secretsmanager':
        access_key = os.environ.get('AWS_SM_SECRET_ACCESS_KEY') or access_key
        secret_key = os.environ.get('AWS_SM_ACCESS_KEY_ID') or secret_key
        sm_endpoint_url = os.environ.get('AWS_SM_ENDPOINT_URL') or endpoint_url
        password_key = os.environ.get('AWS_SM_PASSWORD_KEY') or password_key

        if not all([access_key, secret_key, sm_endpoint_url, password_key]):
            raise ValueError('One or more required environment variables is not set')

        sts_client = boto3.client('sts', aws_access_key_id=access_key, aws_secret_access_key=secret_key)

        response = sts_client.get_session_token()
        access_key_id = response['Credentials']['AccessKeyId']
        secret_access_key = response['Credentials']['SecretAccessKey']
        session_token = response['Credentials']['SessionToken']

        client = boto3.client(
            'secretsmanager',
            endpoint_url=sm_endpoint_url,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            aws_session_token=session_token
        )

        if password_key is not None and password_key != '':
            secret_name = password_key
        try:
            response = client.get_secret_value(SecretId=secret_name)
        except ClientError as e:
            print(f"Error retrieving secret: {e}")
            return None
        else:
            secret_value = response['SecretString']
            supported_encodings = ['utf-8', 'ascii', 'latin-1', 'utf-16']
            if encoding not in supported_encodings:
                print(f"Unsupported encoding: {encoding}")
                return None
            if isinstance(secret_value, bytes):
                secret_value = secret_value.decode(encoding)

        return secret_value
    else:
        raise ValueError(f'invalid password method supplied: "{password_method}".  Valid values include "keyring", "ssm" and "secretsmanager".')


"""Configuration-related utilities"""
#%%
import os
import glob
import configparser
import io

#%%
def read_config_file(config_filepath):
    """Read config .ini file at provided location, using '_local' version instead if found"""
    config = configparser.ConfigParser()
    break_out_flag = False
    # Locate local config file
    local_config_path = ''
    for root, dirs, files in os.walk(os.getcwd()):
        if break_out_flag == True:
            break
        files = [file for file in files if file.endswith(('.ini'))]
        for file in files:
            if break_out_flag == True:
                break
            if file == config_filepath.stem + '_local.ini':
                local_config_path = os.path.join(root, file)
                break_out_flag = True
                break
                #fp_local = io.open(local_config_path, 'r', encoding='utf_8_sig')
    #fp = io.open(config_filepath, 'r', encoding='utf_8_sig')
    config.read([config_filepath, local_config_path])#, encoding='utf-8-sig')
    return config

# %%



"""Datetime-related utilities"""
# from datetime import datetime
import datetime
import math

def get_current_timestamp():
    """Get current timestamp
    current_timestamp: Timestamp object - Used for difference calcs
    filename_timestamp: 'YYYY-MM-DD_HHMMSS' - Used for filenames
    log_timestamp: 'YYYY-MM-DD HH:MM:SS' - Used for logs"""
    
    current_timestamp = datetime.datetime.now()

    filename_timestamp = (
        datetime.datetime
        .fromtimestamp(datetime.datetime.now().timestamp())
        .strftime("%Y-%m-%d_%H%M%S")
        )
    
    log_timestamp = (
        datetime.datetime
        .fromtimestamp(datetime.datetime.now().timestamp())
        .strftime("%Y-%m-%d %H:%M:%S")
        )
                         
    return [current_timestamp, filename_timestamp, log_timestamp]

def getDuration(then, now=datetime.datetime.now()):#, interval = "default"):
    """Return the duration between two timestamps"""
    # Returns a duration as specified by variable interval
    # Functions, except totalDuration, returns [quotient, remainder]

    duration = now - then # For build-in functions
    duration_in_s = duration.total_seconds() 
    
    def get_years():
      return divmod(duration_in_s, 31536000) # Seconds in a year=31536000.

    def get_days(seconds = None):
      return divmod(seconds if seconds != None else duration_in_s, 86400) # Seconds in a day = 86400

    def get_hours(seconds = None):
      return divmod(seconds if seconds != None else duration_in_s, 3600) # Seconds in an hour = 3600

    def get_minutes(seconds = None):
      return divmod(seconds if seconds != None else duration_in_s, 60) # Seconds in a minute = 60

    def get_seconds(seconds = None):
      if seconds != None:
        return divmod(seconds, 1)   
      return duration_in_s

    def total_duration():
        years_list = get_years()
        days_list = get_days(years_list[1]) # Use remainder to calculate next variable
        hours_list = get_hours(days_list[1])
        mins_list = get_minutes(hours_list[1])
        secs_list = get_seconds(mins_list[1])

        return_string = ''

        if int(days_list[0]) > 0:
            return_string += f'{int(days_list[0])} days, '

        if int(hours_list[0]) > 0:
            return_string += f'{int(hours_list[0])} hours, '

        if int(mins_list[0]) > 0:
            return_string += f'{int(mins_list[0])} minutes, '

        return_string += f'{math.ceil(int(secs_list[0]))} seconds'
        
        #return "Time between dates: {} days, {} hours, {} minutes and {} seconds".format(int(d[0]), int(h[0]), int(m[0]), int(s[0]))
        return return_string

    days_between = int(get_days()[0])
    hours_between = int(get_hours()[0])
    minutes_between = int(get_minutes()[0])
    seconds_between = int(get_seconds())
    duration_string = total_duration()
    
    return (
        days_between,
        hours_between,
        minutes_between,
        seconds_between,
        duration_string
    )



from setuptools import setup, find_packages

setup(
    name='nexus-utilities',
    version='0.2.1',
    author='James Larsen',
    author_email='james.larsen42@gmail.com',
    description='Common python utilities',
    long_description='This package is meant to hold various useful utilities for functionality I find myself using across multiple projects.  I will try to keep this documentation updated as I expand the toolkit.',
    long_description_content_type='text/markdown',
    url='https://github.com/james-larsen/nexus-utilities',
    packages=['nexus_utils'],
    install_requires=[
        'boto3>=1.26.45',
        'configparser >=5.3.0',
        'sqlalchemy>=1.4.44',
        'keyring>=23.13.1',
        'chardet>=3.0.4',
        'twine>=3.4.0'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.8',
)
